package org.example;

import io.deephaven.csv.CsvTools;
import io.deephaven.parquet.table.ParquetTools;
import io.deephaven.csv.util.CsvReaderException;
import io.deephaven.engine.context.ExecutionContext;
import io.deephaven.engine.context.TestExecutionContext;
import io.deephaven.engine.table.Table;
import io.deephaven.engine.testutil.ControlledUpdateGraph;
import io.deephaven.enterprise.database.Database;
import io.deephaven.util.SafeCloseable;
import org.junit.jupiter.api.AfterAll;
import org.junit.jupiter.api.Assertions;
import org.junit.jupiter.api.BeforeAll;
import org.junit.jupiter.api.Test;

import static org.mockito.Mockito.mock;
import static org.mockito.Mockito.when;

public class UnitTest {

    private static final String TEST_DATE = "2017-08-25";
    private static Database db;
    private static SafeCloseable executionContextCloseable;

    @BeforeAll
    public static void setup() {
        /*
         * Initialize the execution context for unit tests.
         * Table operations must be run within an open execution context.
         */
        final ExecutionContext executionContext = TestExecutionContext.createForUnitTests();
        final ControlledUpdateGraph updateGraph = (ControlledUpdateGraph) executionContext.getUpdateGraph();
        updateGraph.enableUnitTestMode();
        updateGraph.resetForUnitTests(false);
        executionContextCloseable = executionContext.open();

        db = getTestDatabase();
    }

    @AfterAll
    public static void cleanUp() {
        /*
         * Clean up the execution resources.
         */
        executionContextCloseable.close();
    }

    /**
     * Mocking the database allows us to test our queries' correctness
     * without the complication of creating a real database instance.
     */
    private static Database getTestDatabase() {
        final Database db = mock(Database.class);
        addData(db);
        return db;
    }

    /**
     * The queries we are testing use historical table data. This method sets up the mock database to return
     * the test CSV and Parquet data when db.historicalTable is called.
     */
    private static void addData(final Database db) {
        try {
            final Table t = CsvTools.readCsv("src/test/resources/StockTrades.csv");
            when(db.historicalTable("LearnDeephaven", "StockTrades")).thenReturn(t);
        } catch (CsvReaderException e) {
            throw new RuntimeException("Could not read StockTrades.csv test data", e);
        }

        final Table t = ParquetTools.readTable("src/test/resources/StockQuotes.parquet");
        when(db.historicalTable("LearnDeephaven", "StockQuotes")).thenReturn(t);
    }

    @Test
    public void testAvgPriceQuery() {
        final Table testResult = QueryUtils.getAvgStockPriceTable(db, TEST_DATE);

        Assertions.assertFalse(testResult.isEmpty());

        final double sumAAPL = getFirstDoubleFromColumn(testResult.where("USym = `AAPL`"), "Last");
        Assertions.assertEquals(159.472, sumAAPL, 0.0002);

        final double sumGOOG = getFirstDoubleFromColumn(testResult.where("USym = `GOOG`"), "Last");
        Assertions.assertEquals(923.56, sumGOOG, 0.0002);
    }

    @Test
    public void testMidPriceQuery() {
        final Table testResult = QueryUtils.getMidStockPriceTable(db, TEST_DATE);

        Assertions.assertFalse(testResult.isEmpty());

        final double midPFE = getFirstDoubleFromColumn(testResult.where("USym = `PFE`"), "Mid");
        Assertions.assertEquals(33.285, midPFE, 0.0);
    }

    private static double getFirstDoubleFromColumn(final Table table, final String column) {
        // To extract data from a specific cell in a Table,
        // we first need the row key for the that row position.
        // Row 0 does not necessarily have row key 0.
        final long rowKey = table.getRowSet().get(0);
        // Use the ColumnSource to get the value of the column for that row key
        return table.getColumnSource(column).getDouble(rowKey);
    }
}